#  Copyright (c) 2022 by Amplo.

import numpy as np
import pandas as pd
from sklearn import ensemble
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor

from amplo.utils import get_model


class StackingRegressor:

    _estimator_type = "regressor"

    def __init__(self, **params):
        """
        Wrapper class for Stacking Regressor.

        Parameters
        ----------
        stack list[tuple]: List of tuples (model name, model object)
        params list[dict]: List of model parameters for stack
        """
        # Defaults
        self.trained = False
        self.level_one = None
        self.model = None
        self.params = params
        self.stack = []
        self.n_samples = 0
        self.n_features = 0
        self.mean = None
        self.std = None
        self.set_params(**params)

    def _add_default_models(self, stack: list) -> list:
        """
        Prepares the models stack
        """
        # Add default models
        models = [i[0] for i in stack]
        if "KNeighborsRegressor" not in models:
            stack.append(("KNeighborsRegressor", KNeighborsRegressor()))
        if "DecisionTreeRegressor" not in models:
            stack.append(("DecisionTreeRegressor", DecisionTreeRegressor()))
        if "LinearRegression" not in models:
            stack.append(("LinearRegression", LinearRegression()))
        if "SVR" not in models and self.n_samples < 5000:
            stack.append(("SVR", SVR()))
        return stack

    def fit(self, x: pd.DataFrame, y: pd.Series):
        # Set info
        self.n_samples = x.shape[0]
        self.n_features = x.shape[1]
        self.mean = np.mean(x, axis=0)
        self.std = np.std(x, axis=0)
        self.std[self.std == 0] = 1

        # Normalize
        x = (x - self.mean) / self.std

        # Create stack
        self.level_one = LinearRegression()
        self.stack = self._add_default_models(self.stack)
        self.model = ensemble.StackingRegressor(
            self.stack, final_estimator=self.level_one
        )

        # Fit
        self.model.fit(x, y)

        # Set flag
        self.trained = True

    def set_params(self, **params):
        """
        Set params for the models in the stack

        Parameters
        ----------
        params dict: Nested dictionary, first keys are model names, second params
        """
        # Overwrite old params
        self.params.update(params)

        # Set default
        if "n_samples" in params:
            self.n_samples = params.pop("n_samples")
        if "n_features" in params:
            self.n_features = params.pop("n_features")

        for model_name, param in params.items():
            # Get index
            ind = [i for i, x in enumerate(self.stack) if x[0] == model_name]

            # Add if not in stack
            if len(ind) == 0:
                model = get_model(model_name, mode="regression", samples=self.n_samples)
                self.stack.append((model_name, model.set_params(**param)))

            # Update otherwise
            else:
                self.stack[ind[0]][1].set_params(**param)
        return self

    def get_params(self, **args):
        """
        Returns a dictionary with all params.
        """
        return self.params

    def predict(self, x, *args, **kwargs):
        assert self.trained
        return self.model.predict((x - self.mean) / self.std, *args, **kwargs).reshape(
            -1
        )



#  Copyright (c) 2022 by Amplo.

import json
import os
from pathlib import Path

import joblib
import numpy as np
import pytest
from sklearn.metrics import log_loss, r2_score

from amplo import Pipeline
from amplo.utils.testing import make_interval_data
from tests import make_data, rmtree


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_mode(request, target="target"):
    mode = request.param
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = make_data(mode=mode, target=target)
    yield


@pytest.mark.usefixtures("make_mode")
class TestPipelineByMode:
    def test_mode_detector(self):
        # Numerical test
        pipeline = Pipeline(no_dirs=True, target=self.target)
        pipeline._read_data(self.data)._mode_detector()
        assert pipeline.mode == self.mode

        # Categorical test
        if self.mode == "classification":
            # Convert to categorical
            data = self.data.copy()
            data[self.target] = [f"Class_{v}" for v in data[self.target].values]
            # Detect mode
            pipeline = Pipeline(no_dirs=True, target=self.target)
            pipeline._read_data(data)._mode_detector()
            assert pipeline.mode == self.mode

    def test_stacking(self):
        # Set stacking model
        if self.mode == "classification":
            stacking_model = "StackingClassifier"
        else:
            stacking_model = "StackingRegressor"
        # Fit pipeline
        pipeline = Pipeline(
            target=self.target,
            grid_search_type=None,
            stacking=True,
            extract_features=False,
        )
        pipeline.fit(self.data, model=stacking_model)

    def test_dir_and_settings(self):
        pipeline = Pipeline(
            target=self.target,
            mode=self.mode,
            objective="r2" if self.mode == "regression" else "neg_log_loss",
            n_grid_searches=0,
            plot_eda=False,
            process_data=False,
            extract_features=False,
            document_results=False,
        )
        pipeline.fit(self.data)

        # Test Directories
        assert os.path.exists("Auto_ML")
        assert os.path.exists("Auto_ML/Data")
        assert os.path.exists("Auto_ML/Features")
        assert os.path.exists("Auto_ML/Production")
        assert os.path.exists("Auto_ML/Documentation")
        assert os.path.exists("Auto_ML/Results.csv")

        if self.mode == "classification":
            # Pipeline Prediction
            prediction = pipeline.predict_proba(self.data)
            assert log_loss(self.data[self.target], prediction) > -1

        elif self.mode == "regression":
            # Test data handling
            c, _ = pipeline.convert_data(self.data.drop(self.target, axis=1))
            x = pipeline.x[pipeline.settings["features"]]
            y = self.data[self.target]
            assert np.allclose(
                c.values, x.values
            ), f"Inconsistent X: max diff: {np.max(abs(c.values - x.values)):.2e}"
            assert np.allclose(y, pipeline.y), "Inconsistent y-data"

            # Pipeline Prediction
            prediction = pipeline.predict(self.data)
            assert len(prediction.shape) == 1
            assert r2_score(self.data[self.target], prediction) > 0.75

        else:
            raise ValueError(f"Invalid mode {self.mode}")

        # Settings prediction
        settings = json.load(open("Auto_ML/Production/v1/Settings.json", "r"))
        model = joblib.load("Auto_ML/Production/v1/Model.joblib")
        p = Pipeline(no_dirs=True)
        p.load_settings(settings)
        p.load_model(model)
        if self.mode == "classification":
            assert np.allclose(p.predict_proba(self.data), prediction)
        else:
            assert np.allclose(p.predict(self.data), prediction)

        # Check settings
        assert "pipeline" in settings
        assert "version" in settings
        assert "model" in settings
        assert "amplo_version" in settings
        assert "params" in settings
        assert "drift_detector" in settings
        assert "features" in settings
        assert "validation" in settings
        assert "data_processing" in settings


class TestPipeline:
    def test_drift(self):
        # todo connect with logger
        pytest.skip("This test is not yet implemented")

    def test_all_models(self):
        # TODO
        pytest.skip("This test is not yet implemented")

    def test_interval_analyzer(self):
        """
        Use interval analyzer when a folder with logs is presented to pipeline.
        """
        # Create dummy data
        data_dir = "./test_data"
        make_interval_data(directory=data_dir)

        # Pipeline
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(data_dir)

        # Check if IntervalAnalyser is fitted
        assert pipeline.interval_analyser.is_fitted, "IntervalAnalyser was not fitted"

        # Check data handling
        assert Path(
            "Auto_ML/Data/Interval_Analyzed_v1.csv"
        ).exists(), "Interval-analyzed data was not properly stored"
        assert list(pipeline.data.index.names) == ["log", "index"], "Index is incorrect"
        # TODO: add checks for settings

        # Remove dummy data
        rmtree(data_dir, must_exist=True)



#  Copyright (c) 2022 by Amplo.

"""
Observer for checking production readiness.
"""

from typing import List

from amplo.observation._base import BaseObserver, PipelineObserver
from amplo.observation.data import DataObserver
from amplo.observation.model import ModelObserver

__all__ = ["ProductionObserver"]


class ProductionObserver(BaseObserver):
    """
    Observer before putting to production.

    Testing and monitoring are key considerations for ensuring the production-
    readiness of an ML system, and for reducing technical debt of ML systems.

    Parameters
    ----------
    pipeline : Pipeline
        The amplo pipeline object that will be observed.
    """

    def __init__(self, pipeline):
        super().__init__()

        # Set observers
        self._observers: List[PipelineObserver] = [
            DataObserver(pipeline=pipeline),
            ModelObserver(pipeline=pipeline),
        ]

    def observe(self):
        for obs in self._observers:
            # Make observation
            obs.observe()
            # Gather observations
            self.observations.extend(obs.observations)
